#!/usr/bin/env python

import dbus
import dbus.service
import gobject
from dbus.mainloop.glib import DBusGMainLoop

from syslogger import NullLogger

logger = NullLogger("", 0)

DBusGMainLoop(set_as_default=True)

class ConnectionState:
    UNKNOWN = 0
    PREPARE = 1
    NEED_AUTH = 2
    CONNECT = 3
    IP_CONFIG_GET = 4
    ACTIVATED = 5
    FAILED = 6
    DISCONNECTED = 7

class ServiceState:
    UNKNOWN = 0
    INIT = 1
    SHUTDOWN = 2
    STARTING = 3
    STARTED = 4
    STOPPING = 5
    STOPPED = 6

class PluginFailure:
    LOGIN_FAILED = 0
    CONNECT_FAILED = 1
    BAD_IP_CONFIG = 2

class StartingInProgressError(dbus.DBusException):
    _dbus_error_name = 'org.freedesktop.NetworkManager.VPN.Error.StartingInProgress'

    
class AlreadyStartedError(dbus.DBusException):
    _dbus_error_name = 'org.freedesktop.NetworkManager.VPN.Error.AlreadyStarted'


class StoppingInProgressError(dbus.DBusException):
    _dbus_error_name = 'org.freedesktop.NetworkManager.VPN.Error.StoppingInProgress'

    
class BadArgumentsError(dbus.DBusException):
    _dbus_error_name = 'org.freedesktop.NetworkManager.VPN.Error.BadArguments'

    
class LaunchFailedError(dbus.DBusException):
    _dbus_error_name = 'org.freedesktop.NetworkManager.VPN.Error.LaunchFailed'


class ConnectionInvalidError(dbus.DBusException):
    _dbus_error_name = 'org.freedesktop.NetworkManager.VPN.Error.ConnectionInvalid'


class AlreadyStoppedError(dbus.DBusException):
    _dbus_error_name = 'org.freedesktop.NetworkManager.VPN.Error.AlreadyStopped'

    
class DBusInvalidArgsError(dbus.DBusException):
    _dbus_error_name = 'org.freedesktop.DBus.Error.InvalidArgs'

    
class DBusUnknownInterfaceError(dbus.DBusException):
    _dbus_error_name = 'org.freedesktop.DBus.Error.UnknownInterface'


class DBusFailedError(dbus.DBusException):
    _dbus_error_name = 'org.freedesktop.DBus.Error.Failed'
    
                
class NmVpnPluginService(dbus.service.Object):
    def __init__(self, service_name):
        try:
            name = dbus.service.BusName(service_name, dbus.SystemBus())
            dbus.service.Object.__init__(self, bus_name = name, \
                                           object_path = '/org/freedesktop/NetworkManager/VPN/Plugin')
            #TODO: remove previus object from bus (is it needed?)
            self._loop = gobject.MainLoop()
            self._loop.run()
        except Exception, e:
            logger.error(str(e))        
        # init properties
        self.vpn_plugin_props_read = ('State',)
        self.vpn_plugin_props_write = ()
        self.vpn_plugin_props = {}
        
    def set_state(self, state):
        self.vpn_plugin_props['State'] = state
        self.StateChanged(state)
    
    @dbus.service.method("org.freedesktop.NetworkManager.VPN.Plugin",
                        in_signature='a{sa{sv}}', out_signature='')
    def Connect(self, connection):
        #@note: Tells the plugin to connect
        #@param connection: Describes the connection to be established.
        #@raise StartingInProgressError: 
        #@raise AlreadyStartedError: 
        #@raise StoppingInProgressError: 
        #@raise BadArgumentsError: 
        #@raise LaunchFailedError: 
        logger.debug("Call method Connect: connection = %s" % str(connection))
        self.connect(connection)
    
    @dbus.service.method("org.freedesktop.NetworkManager.VPN.Plugin",
                        in_signature='a{sa{sv}}', out_signature='s')
    def NeedSecrets(self, connection):
        #@note: Asks the plugin whether the provided connection will
        #@note: require secrets to connect successfully.
        #@param connection: Describes the connection that may need secrets.
        #@raise ConnectionInvalidError: 
        #@return: The setting name within the provided connection that requires secrets, if any.
        logger.debug("Call method NeedSecrets: connection = %s" % str(connection))
        return self.need_secrets(connection)
    
    @dbus.service.method("org.freedesktop.NetworkManager.VPN.Plugin",
                        in_signature='', out_signature='')
    def Disconnect(self):
        #@note: Disconnect the plugin.
        #@raise AlreadyStoppedError: 
        #@raise StoppingInProgressError: 
        logger.debug("Call method Disconnect")
        self.disconnect()
        #at the end quiting gobjectmainloop
        self._loop.quit()
    
    @dbus.service.method("org.freedesktop.NetworkManager.VPN.Plugin",
                        in_signature='a{sv}', out_signature='')
    def SetIp4Config(self, config):
        #@note: Set IPv4 details on the connection.
        #@param config: Ip4Config details for the connection
        logger.debug("Call method SetIp4Config: config = %s" % str(config))
        pass
    
    @dbus.service.method("org.freedesktop.NetworkManager.VPN.Plugin",
                        in_signature='s', out_signature='')
    def SetFailure(self, reason):
        #@note: Indicate a failure to the plugin.
        #@param reason: The reason for the failure.
        logger.debug("Call method SetFailure: reason = %s" % str(reason))
        pass
        
    @dbus.service.signal(dbus_interface="org.freedesktop.NetworkManager.VPN.Plugin",
                         signature='u')
    def StateChanged(self, state):
        #@note: Emitted when the plugin state changes.
        #@param state: The new state of the plugin.
        #logger.debug("Call event StateChanged: state = %s" % str(state))
        pass
    
    @dbus.service.signal(dbus_interface="org.freedesktop.NetworkManager.VPN.Plugin",
                         signature='a{sv}')
    def Ip4Config(self, ip4config):
        #@note: The plugin obtained an IPv4 configuration.
        #@param ip4config: The IPv4 configuration
        logger.debug("Call signal Ip4Config: ip4config = %s" % str(ip4config))
        pass
    
    @dbus.service.signal(dbus_interface="org.freedesktop.NetworkManager.VPN.Plugin",
                         signature='s')
    def LoginBanner(self, banner):
        #@note: Emitted when the plugin receives a login banner from the VPN service
        #@param banner: The login banner string
        logger.debug("Call event LoginBanner: banner = %s" % str(banner))
        pass
    
    @dbus.service.signal(dbus_interface="org.freedesktop.NetworkManager.VPN.Plugin",
                         signature='u')
    def Failure(self, reason):
        #@note: Emitted when a failure in the VPN plugin occurs
        #@param reason: Reason code for the failure (NMVpnPluginFailure)
        logger.debug("Call event Failure: reason = %s" % str(reason))
        pass
    
    @dbus.service.method("org.freedesktop.DBus.Properties",
                         in_signature="ss", out_signature="v")
    def Get(self, interface_name, property_name):
        #@return: property value
        logger.debug("Get: interface_name = %s; property_name = %S" % (interface_name, property_name))
        if interface_name == "org.freedesktop.NetworkManager.VPN.Plugin":
            if property_name in self.vpn_plugin_props_read:
                return self.vpn_plugin_props[property_name]
            elif property_name in self.vpn_plugin_props_write:
                raise DBusFailedError('Property %s is not readable' % property_name)
            else:
                raise DBusInvalidArgsError('No such property: %s' % property_name)
        else:
            raise DBusUnknownInterfaceError('Unknown interaface: %s' % interface_name)
    
    @dbus.service.method("org.freedesktop.DBus.Properties",
                         in_signature="ssv", out_signature="")
    def Set(self, interface_name, property_name, value):
        logger.debug("Set: interface_name = %s; property_name = %S; value = %s" % (interface_name, property_name, value))
        if interface_name == "org.freedesktop.NetworkManager.VPN.Plugin":
            if property_name in self.vpn_plugin_props_write:
                self.vpn_plugin_props[property_name] = value
            elif property_name in self.vpn_plugin_props_read:
                raise DBusFailedError('Property %s is not writeable' % property_name)
            else:
                raise DBusInvalidArgsError('No such property: %s' % property_name)
        else:
            raise DBusUnknownInterfaceError('Unknown interaface: %s' % interface_name)

    @dbus.service.method("org.freedesktop.DBus.Properties",
                         in_signature="s", out_signature="a{sv}")
    def GetAll(self, interface_name):
        #@return: properties values
        logger.debug("GetAll: interface_name = %s" % interface_name)
        if interface_name == "org.freedesktop.NetworkManager.VPN.Plugin":
            res = {}
            for k, v in self.vpn_plugin_props:
                if k in self.vpn_plugin_props_read:
                    res[k] = v
            return res
        else:
            raise DBusUnknownInterfaceError('Unknown interaface: %s' % interface_name)
    
    def connect(self, connection):
        #@note: Called from self.Connect. Need to be implemented in child-class.
        pass
    
    def disconnect(self):
        #@note: Called from self.Disconnect. Need to be implemented in child-class.
        pass
        
    def need_secrets(self, connection):
        #@note: Called from self.NeedSecrets. Need to be implemented in child-class.
        return ""

