#!/usr/bin/env python

import re
import os

#TODO: interpriting and writing secrets


class Section(object):
    def __init__(self, type_name, section_name):
        self.type_name = type_name
        self.section_name = section_name
        self.params = {}
        self.alsoflips = {}    
        self.lines = ["%s %s\n" % (self.type_name, self.section_name)]
        
    def set_changed_handler(self, handler):
        self.changed_handler = handler
    
    def param_value_to_line(self, param, value):
        # if value does not contain quotes, but whitespaces
        if re.compile(r'[\s]*^".*"[\s]*$').search(value) == None and \
           re.compile(r'[\s]+').search(value) != None:
           value = '"%s"' % value # quote it
        return "\t\t%s=%s" % (param, value)
        
    def __len__(self):
        return len(self.params)
    
    def __getitem__(self, key):
        if key in self.params:
            return self.params[key][0]
        else:
            raise KeyError("No such key: %s" % key)
    
    def __setitem__(self, key, value):
        if key in self.params:
            linenum = self.params[key][1]
            self.params[key][0] = value
        else:
            linenum = len(self.lines)
            self.params[key] = [value, linenum]
            self.lines += [None]
        self.lines[linenum] = self.param_value_to_line(key, value)
    
    def __delitem__(self, key):
        self.lines[self.params[key][1]] = None
        del self.params[key]
    
    def keys(self):
        return self.params.keys()
    
    def has_key(self, key):
        return self.params.has_key(key)
        
    def alsoflip(self, value):
        if value in self.alsoflips:
            return
        linenum = len(self.lines)
        self.alsoflips[value] = linenum
        self.lines += [self.param_value_to_line('alsoflip', value)]
    
    def rem_alsoflip(self, value):
        if value in self.alsoflips:
            self.lines[self.alsoflips[value]] = None
            del self.alsoflips[value]
    
    def _is_whiteline(self, line):
        return re.compile('^\s*$').search(line) != None
        
    def __str__(self):
        res = ""
        line = ""
        for i in xrange(0, len(self.lines)):
            line = self.lines[i]
            if line == None:
                continue
            if line[-1] != '\n':
                line += '\n'
            res += line
        if not self._is_whiteline(line):
            res += '\n'
        return res

def load_section(lines):
    section = Section('', '')
    # interpriting...
    # interprite header
    exp = re.compile(r'(^conn\s+(?P<connname>.+$))|' + \
                     r'(^config\s+(?P<configname>.+$))')
    mo = exp.match(lines[0])
    if mo == None:
        raise ValueError("Lines is not valid section. Section must start from header")
    gd = mo.groupdict()
    if gd['connname'] != None:
        section.type_name = 'conn'
        section.section_name = gd['connname']
    elif gd['configname'] != None:
        section.type_name = 'config'
        section.section_name = gd['configname']
    else:
        raise ValueError("Unknown error while interpriting section header")
    # interprite content
    exp = re.compile(r'(\s+(?P<parameter>[a-zA-Z0-9\-_.]+)\s*=\s*(("(?P<valuequoted>.+)")|' + \
                     r'(?P<value>[a-zA-Z0-9\-_.]+)))')
    for i in xrange(1, len(lines)): 
        line = lines[i]
        mo = exp.match(line)
        if mo == None:
            continue
        gd = mo.groupdict()
        if gd['value'] != None:
            value = gd['value']
        else:
            value = gd['valuequoted']
        if gd['parameter'] == 'alsoflip':
            section.alsoflips[value] = i
        else:
            section.params[gd['parameter']] = [value, i]        
    section.lines = lines
    return section


class IpsecConf(object):
    UNKNOWN = 0
    COMMENT = 1
    SECTION = 2
    INCLUDE = 3
    WHITELINE = 4
    NONE    = 5
    
    def __init__(self, path):
        self.path = path
        self.includes = {}
        self.sections = {}
        self.logic = []
        #compile regular expression
        exp = re.compile(r'(^[#]+(?P<comment>.*$))|' + \
                         r'(^include\s+(?P<include>.+$))|' + \
                         r'(^config\s+(?P<config_name>.*$))|' + \
                         r'(^conn\s+(?P<conn_name>.*$))|' + \
                         r'(^(?P<whiteline>\s*$))|' + \
                         r'(\s+(?P<section>.*$))')
        if not os.path.exists(path):
            return
        f = open(path)
        section_lines = []
        for line in f:
            mo = exp.search(line)
            if mo == None:
                if len(section_lines) > 0:
                    self._create_and_add_section(section_lines)
                    section_lines = []
                self.logic += [[IpsecConf.UNKNOWN, line]]
                continue
            gd = mo.groupdict()
            if gd['comment'] != None:
                if len(section_lines) > 0:
                    self._create_and_add_section(section_lines)
                    section_lines = []
                self.logic += [[IpsecConf.COMMENT, line]]
            elif gd['include'] != None:
                if len(section_lines) > 0:
                    self._create_and_add_section(section_lines)
                    section_lines = []
                if gd['include'] not in self.includes:
                    linenum = len(self.logic)
                    self.includes[gd['include']] = linenum
                    self.logic += [[IpsecConf.INCLUDE, line]]
            elif gd['config_name'] != None or gd['conn_name'] != None:
                if len(section_lines) > 0:
                    self._create_and_add_section(section_lines)
                    section_lines = []
                section_lines += [line]
            elif gd['whiteline'] != None:
                if len(section_lines) > 0:
                    section_lines += [line]
                else:
                    self.logic += [[IpsecConf.WHITELINE, line]]                
            elif gd['section'] != None:
                if len(section_lines) > 0:
                    section_lines += [line]
                else:
                    self.logic += [[IpsecConf.UNKNOWN, line]]
        f.close()
        if len(section_lines) > 0:
            self._create_and_add_section(section_lines)
            section_lines = []
        pass
    
    def _create_and_add_section(self, section_lines):
        linenum = len(self.logic)
        section = load_section(section_lines)
        self.sections[section.section_name] = [section, linenum]
        self.logic += [[IpsecConf.SECTION, section]]        
            
    def include(self, path):
        if path in self.includes:
            return
        linenum = len(self.logic)
        self.includes[path] = linenum
        self.logic += [[IpsecConf.INCLUDE, 'include %s\n' % path]]
    
    def rem_include(self, path):
        if path not in self.includes:
            raise KeyError("No such include: %s" % path)
        self.logic[self.includes[path]] = [IpsecConf.NONE, None]
        del self.includes[path]
    
    def get_includes(self):
        #return: list of includes
        return self.includes.keys()
        
    def add_section(self, section):        
        linenum = len(self.logic)
        self.sections[section.section_name] = [section, linenum]
        self.logic += [[IpsecConf.SECTION, section]]
    
    def rem_section(self, section_name):
        if section_name not in self.sections:
            raise KeyError("No such section: %s" % section_name)
        linenum = self.sections[section_name][1]
        del self.sections[section_name]
        self.logic[linenum] = [IpsecConf.NONE, None]
    
    def get_section(self, section_name):
        return self.sections[section_name][0]
    
    def save(self):
        f = open(self.path, "wt")
        prev = IpsecConf.NONE
        for i in xrange(0, len(self.logic)):
            ltype, content = self.logic[i]
            if ltype == IpsecConf.NONE:
                continue
            #if prev == IpsecConf.SECTION and \
            #   ltype != IpsecConf.WHITELINE:
            #    f.write('\n')
            f.write(str(content))
            prev = ltype
        f.close()
    
    def clear(self):
        self.includes = {}
        self.sections = {}
        self.logic = []
        
    def __str__(self):
        return str(self.logic)

#parse only includes
class IpsecSecrets(object):
    def __init__(self, path):
        self.path = path
        self.includes = {}
        self.lines = []
        if not os.path.exists(path):
            return
        f = open(path)
        exp = re.compile(r'^include\s+(?P<include>.+)$')        
        for line in f:
            mo = exp.match(line)
            self.lines += [line]
            if mo == None:
                continue
            gd = mo.groupdict()
            if gd['include'] == None:
                continue
            self.includes[gd['include']] = len(self.lines) - 1
        f.close()            

    def include(self, path):
        if path in self.includes:
            return
        linenum = len(self.includes)
        self.includes[path] = linenum
        self.lines += ['include %s\n' % path]
    
    def rem_include(self, path):
        if path not in self.includes:
            raise KeyError("No such include: %s" % path)
        self.lines[self.includes[path]] = None
        del self.includes[path]
    
    def get_includes(self):
        #return: list of includes
        return self.includes.keys()
        
    def add_rsa(self, hosts, key):
        line = '%s: rsa {%s}\n' % (" ".join(hosts), key)
        self.lines += [line]
        
    def add_x509(self, path, passphrase = '', promt = False):
        if len(passphrase) > 0:
            line = ':RSA "%s" "%s"\n' % (path, passphrase)
        elif promt:
            line = ':RSA "%s" %%promt\n' % path
        else:
            line = ':RSA "%s"\n' % path
        self.lines += [line]
        
    def add_psk(self, hosts, passphrase):
        line = '%s: PSK "%s"\n' % (" ".join(hosts), passphrase)
        self.lines += [line]
    
    def save(self):
        f = open(self.path, "wt")
        for i in xrange(0, len(self.lines)):
            el = self.lines[i]
            if el == None:
                continue
            f.write(el)
        f.close()
    
    def clear(self):
        self.lines = []
        self.includes = {}

import subprocess 

from syslogger import SysLogger
from threading import Thread

logger = SysLogger("NMVPNOpenswanService", 0)

class ConnectionException(Exception):
    pass

#TODO: create decorators for async methods, if it will be needed more then once
class Connection(object):
    def __init__(self, name):
        self.name = name

    def start_async(self, callback = None):
        #param callback: func(error)
        thread = Thread(target = self._start, args = (callback,))
        thread.setDaemon(True)
        thread.start()      
    
    def _start(self, callback = None):
        p = subprocess.Popen(["ipsec", "auto", "--add", self.name], stdout = subprocess.PIPE)
        p.wait()
        addresult = ";".join(p.stdout.readlines())
        res = p.poll()
        logger.debug("Result of 'ipsec auto --add %s' = %s" % (self.name, res))
        if not res and len(addresult) == 0:
            logger.debug("Run ipsec auto --up %s" % self.name)
            p = subprocess.Popen(["ipsec", "auto", "--up", self.name], stdout = subprocess.PIPE)
            res = None
        else:
            res = ConnectionException("Couldn't add connection %s: %s" % (self.name, addresult))
        if callback != None:
            callback(res)
        else:
            return res
        
    def start(self):
        ret = self._start()
        if ret != None:
            raise ret

    def stop(self):
        p = subprocess.Popen(["ipsec", "auto", "--down", self.name])
        p.wait()    
        p = subprocess.Popen(["ipsec", "auto", "--delete", self.name])
        p.wait()
        
class IpsecControl(object):
    def __init__(self):
        pass
        
    def start(self):
        p = subprocess.Popen(["ipsec", "setup", "start"])
        p.wait()

